// FFTEST.cpp : Defines the entry point for the console application.
//
#define _WIN32_WINNT 0x0501

#include <windows.h>
#include <string.h>
#include <stdio.h>
#include <aclapi.h>
#include <assert.h>
#include "..\sdk\plugin.h"
#include "..\sdk\takecmd.h"

DWORD WINAPI EventThread( LPVOID lpvParam );
int GetHandleList( LPTSTR lpszString, HANDLE *hEvt, int nMaxHandles );
void usage( void );

#define HANDLE_LIMIT	MAXIMUM_WAIT_OBJECTS 
struct evt_struct
{
	HANDLE hEvt[HANDLE_LIMIT];
	DWORD  timeout;
	int	   nHandles;
	BOOL   fWaitAll;
};

struct evt_handle
{
	HANDLE	hEvt;
	BOOL	bValid;
};

#define HANDLE_TO_INDEX(x)	(x + 1000)
#define INDEX_TO_HANDLE(x)	(x - 1000)

SECURITY_DESCRIPTOR sd;
SECURITY_ATTRIBUTES sa;
HANDLE hThread;
evt_handle	HandleSet[ HANDLE_LIMIT ];

BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_PROCESS_DETACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
			break;
	}
	return TRUE;
}

DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;
#ifdef _WIN64
#pragma message( "Win64 Configuration" )
	piInfo.pszDll = (LPTSTR)L"Event64";
#else
#pragma message( "Win32 Configuration" )
	piInfo.pszDll = (LPTSTR)L"Event";
#endif
	piInfo.pszAuthor = (LPTSTR)L"Scott A. Mintz";
	piInfo.pszEmail = NULL;
	piInfo.pszWWW = NULL;
	piInfo.pszDescription = (LPTSTR)L"Wait for or signal an event\r\n\r\n"
							L"@EventCreate[[event_name]]\r\n"
							L"\tcreate an event.  If event_name is not spec'd then it is an unnamed event\r\n"
							L"\tand can only be used by this instance of TCC.\r\n"
							L"@MutexCreate[[mutex_name]]\r\n"
							L"\tcreate a mutex.  If mutex_name is not spec'd then it is an unnamed mutex\r\n"
							L"\tand can only be used by this instance of TCC.\r\n"
							L"@EventClose[handle]\r\n"
							L"\tclose the handle of the event created with @EventCreate\r\n"
							L"@MutexClose[handle]\r\n"
							L"\tclose the handle of the mutex created with @MutexCreate\r\n"
							L"EventReset ev1 [ev2 ... [ev64]]\r\n"
							L"\tset the specified event object(s) to the non-signaled state\r\n"
							L"EventSignal ev1 [ev2 ... [ev64]]\r\n"
							L"\tset the specified event object(s) to the signaled state\r\n"
							L"EventWait [/t timeout_ms] [/A(ll)] ev1 [ev2 ... [ev64]]\r\n"
							L"\twait for Any of the specified events/mutexes with an optional timeout.\r\n"
							L"\t/A - wait for All of the events\r\n"
							L"MutexRelease ev1 [ev2 ... [ev64]]\r\n"
							L"\treleases ownership of the specified mutex object(s).\r\n";
	piInfo.pszFunctions = (LPTSTR)L"EventSignal,EventReset,EventWait,MutexRelease,@EventClose,@EventCreate,@MutexCreate,@MutexClose";
	piInfo.nMajor = 2;
	piInfo.nMinor = 0;
	piInfo.nBuild = 1;

	return &piInfo;
}


// __declspec(dllexport) BOOL _stdcall InitializePlugin( void )
DLLExports BOOL WINAPI InitializePlugin( void )
{
    // Initialize a security descriptor.  
    if (!InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION)) 
    {  
		Qprintf(GetStdHandle( STD_ERROR_HANDLE ), L"InitializeSecurityDescriptor Error %u\n\r", GetLastError());
        return TRUE;
    } 
 
    // Add the ACL to the security descriptor. 
    if (!SetSecurityDescriptorDacl(&sd, 
            TRUE,     // bDaclPresent flag   
            NULL, 
            FALSE))   // not a default DACL 
    {  
        Qprintf(GetStdHandle( STD_ERROR_HANDLE ), L"SetSecurityDescriptorDacl Error %u\n\r", GetLastError());
        return TRUE;
    } 

    // Initialize a security attributes structure.
    sa.nLength = sizeof (SECURITY_ATTRIBUTES);
    sa.lpSecurityDescriptor = &sd;
    sa.bInheritHandle = TRUE;

	//initialize the handle array
	for(int i=0; i< HANDLE_LIMIT; i++ )
		HandleSet[i].bValid = FALSE;

	return 0;
}


DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	// close any event handles that were left open
	for(int i=0; i< HANDLE_LIMIT; i++ )
		if( HandleSet[i].bValid ) CloseHandle( HandleSet[i].hEvt );
	return 0;
}

DLLExports INT WINAPI f_EventCreate( LPTSTR lpszString )
{
	int i;

	for( i=0; i< HANDLE_LIMIT; i++ )
	{
		if( !HandleSet[i].bValid )
			break;
	}

	if( i == HANDLE_LIMIT )
	{
		TCError( ERROR_NOT_ENOUGH_MEMORY, L"Event Handle Limit has been reached." );
		return 0;
	}

	HANDLE hEvt = CreateEvent( &sa, TRUE, FALSE, lpszString );
	if(hEvt == NULL )
		return 1;

	HandleSet[i].hEvt = hEvt;
	HandleSet[i].bValid = TRUE;
	Sprintf( lpszString, L"%d", HANDLE_TO_INDEX(i) );

	return 0;
}

DLLExports INT WINAPI f_MutexCreate( LPTSTR lpszString )
{
	int i;

	for( i=0; i< HANDLE_LIMIT; i++ )
	{
		if( !HandleSet[i].bValid )
			break;
	}

	if( i == HANDLE_LIMIT )
	{
		TCError( ERROR_NOT_ENOUGH_MEMORY, L"Event Handle Limit has been reached." );
		return 0;
	}

	HANDLE hEvt = CreateMutex( &sa, FALSE, lpszString );
	if(hEvt == NULL )
		return 1;

	HandleSet[i].hEvt = hEvt;
	HandleSet[i].bValid = TRUE;
	Sprintf( lpszString, L"%d", HANDLE_TO_INDEX(i) );

	return 0;
}

DLLExports INT WINAPI f_EventClose( LPTSTR lpszString )
{
	unsigned int i;

	if ( lpszString == NULL )
		return 1;

	Sscanf( lpszString, L"%d", &i );

	i = INDEX_TO_HANDLE(i);
	if( i >= HANDLE_LIMIT || !HandleSet[i].bValid )
	{
		Sprintf( lpszString, L"%%[%d]", HANDLE_TO_INDEX(i) );
		TCError( ERROR_INVALID_HANDLE, lpszString );
		return 0;
	}

	HandleSet[i].bValid = FALSE;
	if( CloseHandle( HandleSet[i].hEvt ) )
		return 0;
	else
		return 1;
}

DLLExports INT WINAPI EventSignal(LPTSTR lpszString)
{
	if ( lpszString == NULL )
		return 1;

	HANDLE hEvt[HANDLE_LIMIT];
	int nSuccesses = 0;
	int nHandles = GetHandleList( lpszString, hEvt, HANDLE_LIMIT );
	if( nHandles < 0 )
	{
		Qprintf( GetStdHandle(STD_ERROR_HANDLE), 
				L"EventSignal - sets the specified event object(s) to the signaled state.\n\r\n\r"
				L"EventSignal ev1 [ev2 ... [ev64]]\r\n"
				L"	ev1 [ev2 ... [ev100]] - handles returned by @EventCreate\n\r" );
		return -1;
	}

	for( int i=0; i< nHandles; i++ )
	{
		SetEvent( hEvt[i] );
	}
	return nHandles ? 0 : 1;
}

DLLExports INT WINAPI EventReset(LPTSTR lpszString)
{
	if ( lpszString == NULL )
		return 1;

	HANDLE hEvt[HANDLE_LIMIT];
	int nSuccesses = 0;
	int nHandles = GetHandleList( lpszString, hEvt, HANDLE_LIMIT );
	if( nHandles < 0 )
	{
		Qprintf( GetStdHandle(STD_ERROR_HANDLE), 
				L"EventReset - sets the specified event object(s) to the non-signaled state.\n\r\n\r"
				L"EventReset ev1 [ev2 ... [ev64]]\r\n"
				L"	ev1 [ev2 ... [ev100]] - handles returned by @EventCreate\n\r" );
		return -1;
	}

	for( int i=0; i< nHandles; i++ )
	{
		ResetEvent( hEvt[i] );
	}
	return nHandles ? 0 : 1;
}

DLLExports INT WINAPI MutexRelease(LPTSTR lpszString)
{
	if ( lpszString == NULL )
		return 1;

	HANDLE hEvt[HANDLE_LIMIT];
	int nSuccesses = 0;
	int nHandles = GetHandleList( lpszString, hEvt, HANDLE_LIMIT );
	if( nHandles < 0 )
	{
		Qprintf( GetStdHandle(STD_ERROR_HANDLE), 
				L"MutexRelease - releases ownership of the specified mutex object(s).\n\r\n\r"
				L"MutexRelease ev1 [ev2 ... [ev64]]\r\n"
				L"	ev1 [ev2 ... [ev100]] - handles returned by @MutexCreate\n\r" );
		return -1;
	}

	for( int i=0; i< nHandles; i++ )
	{
		ReleaseMutex( hEvt[i] );
	}
	return nHandles ? 0 : 1;
}

DLLExports INT WINAPI EventWait(LPTSTR lpszString)
{
	evt_struct	sEvt;
	DWORD		ret;
	int			argc;
	LPWSTR		*argv, *argv_org;
	unsigned int i;

	argv = CommandLineToArgvW(lpszString, &argc);
	argv_org = argv;

	if( argc < 1 ) 
	{
		usage();
		LocalFree( argv_org );
		return -2;
	}

	sEvt.nHandles = 0;
	sEvt.fWaitAll = FALSE;
	sEvt.timeout = INFINITE;

	argv++;
	argc--;
	while( argc )
	{
		if( argv[0][0] == L'-' || argv[0][0] == L'/' )
		{
			switch( argv[0][1] )
			{
			case L'A':
			case L'a':
				sEvt.fWaitAll = TRUE;
				break;
			case L't':
			case L'T':
				if( argv[0][2] == L'\0' )
				{
					argv++;
					argc--;
					Sscanf( argv[0], L"%d", &sEvt.timeout );
				}
				else
					Sscanf( &argv[0][2], L"%d", &sEvt.timeout );
				break;
			default:
				usage();
				LocalFree( argv_org );
				return -2;
			}
		}
		else
		{
			if( sEvt.nHandles >= HANDLE_LIMIT )
			{
				Qprintf( GetStdHandle(STD_ERROR_HANDLE), L"Too many handles specified.  Extra ignored.\r\n" );
			}
			else
			{
				Sscanf( argv[0], L"%d", &i );
				i = INDEX_TO_HANDLE(i);
				if( i >= HANDLE_LIMIT || !HandleSet[i].bValid )
				{
					Sprintf( lpszString, L"%%[%d]", HANDLE_TO_INDEX(i) );
					TCError( ERROR_INVALID_HANDLE, lpszString );
					LocalFree( argv_org );
					return -2;
				}
				sEvt.hEvt[sEvt.nHandles++] = HandleSet[i].hEvt;
			}
		}

		argv++;
		argc--;
	}

	// Free the command line args
	LocalFree( argv_org );

	if( !sEvt.nHandles )
	{
		usage();
		return -2;
	}

	do
	{
		ret = WaitForMultipleObjects( sEvt.nHandles, sEvt.hEvt, sEvt.fWaitAll, 100 );
		if( ret == WAIT_TIMEOUT )
		{
			if( sEvt.timeout != INFINITE )
				if( sEvt.timeout > 100 )
					sEvt.timeout -= 100;
				else
					return -1;  //timeout
		}
		else
			return ret - WAIT_OBJECT_0 + 1;

		tty_yield(0);	//let TCMD do something
	}while(1);
}

int GetHandleList( LPTSTR lpszString, HANDLE *hEvt, int nMaxHandles )
{
	int		argc;
	int		nHandles = 0;
	unsigned int i;
	LPWSTR	*argv, *argv_org;

	argv = CommandLineToArgvW(lpszString, &argc);
	argv_org = argv;

	if( argc <= 1 ) 
	{
		LocalFree( argv_org );
		return -1;
	}

	argv++;
	argc--;
	while( argc && nHandles < nMaxHandles )
	{
		if( argv[0][0] == L'?' || argv[0][1] == L'?' )
		{
			nHandles = -1;
			break;
		}
		Sscanf( argv[0], L"%d", &i );
		i = INDEX_TO_HANDLE(i);
		if( i >= HANDLE_LIMIT || !HandleSet[i].bValid )
		{
			Sprintf( lpszString, L"%%[%d]", HANDLE_TO_INDEX(i) );
			TCError( ERROR_INVALID_HANDLE, lpszString );
			nHandles = 0;
			break;
		}
		hEvt[nHandles++] = HandleSet[i].hEvt;

		argv++;
		argc--;
	}

	// Free the command line args
	LocalFree( argv_org );

	return nHandles;
}


void usage( void )
{
	Qprintf( GetStdHandle(STD_ERROR_HANDLE), 
				L"EventWait - wait for any of the specified  events/mutexes\n\r\n\r"
				L"EventWait [/T n] [/A] ev1 [ev2 ... [ev64]]\r\n"
				L"	/T(imeout) in milliseconds\n\r"
				L"  /A(ll) - wait for All of the events\r\n"
				L"	event_handle(s) returned by @EventCreate or @MutexCreate\n\r" );
}

